( cd usr/lib64/gcc/x86_64-slamd64-linux/4.2.4/adalib ; rm -rf libgnarl.so )
( cd usr/lib64/gcc/x86_64-slamd64-linux/4.2.4/adalib ; ln -sf libgnarl-4.2.so libgnarl.so )
( cd usr/lib64/gcc/x86_64-slamd64-linux/4.2.4/adalib ; rm -rf libgnat.so )
( cd usr/lib64/gcc/x86_64-slamd64-linux/4.2.4/adalib ; ln -sf libgnat-4.2.so libgnat.so )
